# -*- coding: utf-8 -*-
from . import Command as _Command
from .. import io

from ...vendor.terraform import TFVarsConfig


class Status(_Command):
    def _validate_and_configure(self):
        super(self.__class__, self)._validate_and_configure()

        self.tf = TFVarsConfig(self.tfvars_path)
        self.tf.load()

    def _run(self):
        services = self.tf.get_services()

        docker_org = self.config['DOCKER_ORG']
        table_data = [
            ['id', 'name', 'image tag (%s)' % docker_org, 'running', 'is active'],
        ]
        for i, service_name in enumerate(sorted(services.iterkeys()), 1):
            artifact = services[service_name]

            image_tag = artifact['image'].split(':')[-1]
            running = artifact['desired_count']
            is_active = 'yes' if int(running) > 0 else 'no'

            table_data.append([i, service_name, image_tag, running, is_active])

        io.info('displaying "%s" active&non-active services on %s' % (docker_org, self.env))
        io.print_table(table_data, 'current %s artifacts' % self.env)
        io.warn('only dockerized services are shown here currently...')



# -*- coding: utf-8 -*-
from datetime import datetime
from dateutil import tz

import humanize

from . import Command as _Command
from .. import io

from ...domain.vendor import NoChangesEmptyCommit
from ...vendor.terraform import TFVarsConfig
from ...vendor import dockerhub, git_extensions


class Bump(_Command):
    def _validate_and_configure(self):
        super(self.__class__, self)._validate_and_configure()

        self.tf = TFVarsConfig(self.tfvars_path)
        self.tf.load()

        self.deployment_repo = self.config['TF_DEPLOYMENT_REPO']

        self.service_name = self.kwargs['service']
        self.service_name_normalized = self.tf.normalize_service_name(self.service_name)

        self.should_push = self.kwargs['should_push']
        self.is_latest = self.kwargs['is_latest']

        self.artifact_key = self.tf.get_artifact_key(self.service_name_normalized)

        io.info('authenticating "%s" against dockerhub' % self.config['DOCKER_ORG'])
        self.auth = dockerhub.DockerHubAuthentication(
            self.config['DOCKER_USERNAME'],
            self.config['DOCKER_PASSWORD'],
            self.config['DOCKER_ORG']
        )

    def _prompt_artifact_selection(self, artifacts):
        current_image = self.tf.get_var(self.artifact_key)

        io.info('found artifacts for "%s/%s"' % (self.config['DOCKER_ORG'], self.service_name))
        table_data = [
            ['id', 'tag name (* = current)', 'created at', 'size'],
        ]
        for i, artifact in enumerate(artifacts, 1):
            last_updated = datetime.strptime(artifact['last_updated'], '%Y-%m-%dT%H:%M:%S.%fZ')
            last_updated = last_updated.replace(tzinfo=tz.gettz('UTC'))
            last_updated = last_updated.astimezone(tz.tzlocal())

            last_updated_friendly = last_updated.strftime('%a %d %b, %I:%M%p')
            last_updated_humanized = humanize.naturaltime(last_updated.replace(tzinfo=None))

            updated_at = '%s (%s)' % (last_updated_friendly, last_updated_humanized)
            image_size = humanize.naturalsize(artifact['full_size'])
            image_name = artifact['name']
            if image_name in current_image:  # indicate the current artifact.
                image_name += ' *'

            table_data.append([i, image_name, updated_at, image_size])
        io.print_table(table_data, 'recent artifacts')

        # Handle the case where the selected artifact is the current artifact.
        selected_artifact = io.collect_input('select the artifact you want to use [q]:', artifacts)
        if selected_artifact and selected_artifact['name'] in current_image:
            io.err('selected artifact is already the current active artifact')
            return None
        return selected_artifact

    def _prompt_latest_confirmation(self, artifacts):
        latest_artifact = artifacts[0]
        tag = latest_artifact['name']

        input_ = io.collect_single_input('are you sure (latest: %s)? [y/n/q]:' % tag)
        if input_ not in ['y', None]:
            return None
        return latest_artifact

    def _get_artifact(self):
        io.info('retrieving image tags for "%s" from dockerhub' % self.service_name)
        artifacts = dockerhub.list_image_tags(self.auth, self.service_name)
        artifacts = artifacts[:10]

        if not artifacts:
            io.err('no artifacts were found for "%s/%s"' % (self.config['DOCKER_ORG'], self.service_name))
            return None

        if self.is_latest:
            return self._prompt_latest_confirmation(artifacts)
        return self._prompt_artifact_selection(artifacts)

    def _run(self):
        artifact = self._get_artifact()
        if artifact is None:  # An artifact wasn't selected, end command.
            return None
        tag = artifact['name']

        image_abspath = dockerhub.build_image_abspath(self.auth, self.service_name_normalized, tag)
        io.info('updating "%s"' % image_abspath)

        self.tf.update_var(self.artifact_key, image_abspath)
        self.tf.save()

        commit_message = 'chore(variables): bumped %s:%s' % (self.service_name, tag)

        did_commit = git_extensions.commit_changes(self.deployment_repo, commit_message)
        if not did_commit:
            raise NoChangesEmptyCommit('"%s" has nothing to commit' % self.deployment_repo.working_dir)

        if self.should_push:
            git_extensions.push_commits(self.deployment_repo)
        else:
            io.warn('commit to tfvars was NOT pushed to remote!')
            io.warn("it's your responsibility to bundle changes and explicitly push")



# -*- coding: utf-8 -*-
from . import BaseMimironException
from git import InvalidGitRepositoryError as _InvalidGitRepositoryError


# PyGit custom exceptions


class InvalidGitRepo(BaseMimironException, _InvalidGitRepositoryError):
    def __init__(self, path):
        super(self.__class__, self).__init__('"%s" is an invalid git repository ' % path)


class FetchRemoteUnknownNextStep(BaseMimironException):
    pass


class UnexpectedGitError(BaseMimironException):
    pass


class NoChangesEmptyCommit(BaseMimironException):
    pass


# Terraform variables custom exceptions


class TFVarsMissingConfigFile(BaseMimironException):
    def __init__(self, path):
        message = 'cannot find terraform tfvars config "%s"' % path
        super(self.__class__, self).__init__(message)


class TFVArsConfigNeverLoaded(BaseMimironException):
    def __init__(self):
        message = 'tfvars config was never loaded'
        super(self.__class__, self).__init__(message)


class InvalidTFVarsConfig(BaseMimironException):
    def __init__(self, path, error):
        message = 'error found in tfvars config "%s"\n %s' % (path, error)
        super(self.__class__, self).__init__(message)


# DockerHub custom exceptions


class InvalidDockerHubCredentials(BaseMimironException):
    def __init__(self):
        message = 'failed to authenticate against dockerhub'
        super(self.__class__, self).__init__(message)


class DockerConnectionError(BaseMimironException):
    def __init__(self):
        message = 'failed to connect to dockerhub auth servers'
        super(self.__class__, self).__init__(message)



# -*- coding: utf-8 -*-
import json
import copy

from ..domain.vendor import TFVarsMissingConfigFile
from ..domain.vendor import InvalidTFVarsConfig
from ..domain.vendor import TFVArsConfigNeverLoaded


class TFVarsConfig(object):
    SUPPORTED_TYPES = {
        'docker': 'DOCKER_IMAGE',
        'ami': 'AMI',
    }

    def __init__(self, tfvars_path):
        self.tfvars_path = tfvars_path

        self._config = {}
        self.has_loaded_config = False

    @property
    def config(self):
        return copy.deepcopy(self._config)

    def load(self):
        try:
            with open(self.tfvars_path, 'rU') as f:
                self._config = json.load(f)
        except IOError:
            raise TFVarsMissingConfigFile(self.tfvars_path)
        except (TypeError, ValueError) as e:
            raise InvalidTFVarsConfig(self.tfvars_path, e)

        self.has_loaded_config = True
        return copy.deepcopy(self._config)

    def save(self):
        if not self.has_loaded_config:
            raise TFVArsConfigNeverLoaded
        try:
            with open(self.tfvars_path, 'w') as f:
                data = json.dumps(self._config, f, ensure_ascii=False, indent=2, sort_keys=True)
                data = data.split('\n')
                data = [d.rstrip() for d in data]
                data = '\n'.join(data)
                f.write(data)
        except IOError:
            raise TFVarsMissingConfigFile(self.tfvars_path)

    def normalize_service_name(self, service):
        return str(service).strip().replace('-', '_')

    def get_artifact_key(self, service, is_ami=False):
        return service + ('_ami_id' if is_ami else '_image')

    def get_services(self):
        config = self.config

        services = {}
        for k in config.iterkeys():
            if k.endswith('_image'):
                services[k.replace('_image', '')] = {}
        for service_name, service_data in services.iteritems():
            for k, v in config.iteritems():
                if k.startswith(service_name):
                    service_data[k.replace(service_name + '_', '')] = v
        return services

    def update_var(self, name, value):
        self._config[name] = value

    def get_var(self, name):
        return self._config[name]



# -*- coding: utf-8 -*-
import requests
import json

from ..domain.vendor import InvalidDockerHubCredentials
from ..domain.vendor import DockerConnectionError


class DockerHubAuthentication(object):
    def __init__(self, username, password, org, generate_token=True):
        self.username = username
        self.password = password
        self.org = org

        self._token = None
        if generate_token:
            self._token = self.generate_token()

    def generate_token(self):
        payload = json.dumps({
            'username': self.username, 'password': self.password,
        })
        headers = {
            'Content-Type': 'application/json',
        }
        endpoint = 'https://hub.docker.com/v2/users/login/'

        try:
            response = requests.post(endpoint, data=payload, headers=headers)
            if response.status_code == 200:
                return response.json()['token']
            raise InvalidDockerHubCredentials
        except requests.exceptions.ConnectionError:
            raise DockerConnectionError

    @property
    def token(self):
        if not self._token:
            self._token = self.generate_token()
        return self._token

    @token.setter
    def token(self, new_token):
        self._token = new_token


def _api_request(endpoint, method, auth):
    token = auth.token
    if token is None:
        return None

    try:
        response = method(endpoint, headers={'Authorization': 'JWT %s' % token})
        return response.json() if response.status_code == 200 else None
    except requests.exceptions.ConnectionError:
        raise DockerConnectionError


def list_repositories(auth, page_size=100):
    endpoint = 'https://hub.docker.com/v2/repositories/%s/?page_size=%s' % (
        auth.org, page_size,
    )
    response = _api_request(endpoint, requests.get, auth)
    return response['results'] if response is not None else response


def list_image_tags(auth, image_name, page_size=100):
    endpoint = 'https://hub.docker.com/v2/repositories/%s/%s/tags/?page_size=%s' % (
        auth.org, image_name, page_size,
    )
    response = _api_request(endpoint, requests.get, auth)
    return response['results'] if response is not None else response


def build_image_abspath(auth, image_name, tag):
    return auth.org + '/' + image_name + ':' + tag



# -*- coding: utf-8 -*-
from functools import wraps

import os

from git.exc import GitCommandError

from ..domain.vendor import UnexpectedGitError
from ..domain.vendor import FetchRemoteUnknownNextStep
from ..core import io


def git_failure(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except GitCommandError as e:
            raise UnexpectedGitError(e)
    return wrapper


@git_failure
def get_ahead_behind_count(repo):
    branch = repo.active_branch

    commits_behind = repo.iter_commits('%s..origin/%s' % (branch, branch))
    commits_behind = sum(1 for _ in commits_behind)

    commits_ahead = repo.iter_commits('origin/%s..%s' % (branch, branch))
    commits_ahead = sum(1 for _ in commits_ahead)

    return commits_ahead, commits_behind


@git_failure
def sync_updates(repo, push=False):
    """Updates the local `repo` with origin, pulling and push when necessary.

    True is returned when any changes were made (aside fetch), otherwise False.
    """
    ref = repo.remotes.origin.refs[0].name
    repo_dir = repo.working_dir
    branch = repo.active_branch

    io.info('fetching remote for "%s"@:%s' % (os.path.split(repo_dir)[-1], branch))
    repo.remotes.origin.fetch()

    ahead, behind = get_ahead_behind_count(repo)
    is_dirty = repo.is_dirty()

    if not ahead and not behind and not is_dirty:
        return False

    io.info('(found) [ahead: %s, behind: %s] [dirty: %s]' % (ahead, behind, is_dirty))

    # possible merge conflicts
    if (ahead and behind) or (behind and is_dirty):
        raise FetchRemoteUnknownNextStep('possible merge conflict. please manually resolve')

    # we're ahead so let's push these changes up
    if ahead and push:
        io.warn('ahead, pushing local changes to %s' % ref)
        repo.remotes.origin.push()

    # we're behind so let's pull changes down
    if behind:
        io.warn('behind, pulling changes from %s' % ref)
        repo.remotes.origin.pull()
    return True


@git_failure
def get_recent_commits(repo, limit=10):
    return list(repo.iter_commits(max_count=limit))


@git_failure
def commit_changes(repo, commit_message):
    if not repo.is_dirty():
        return False

    repo.git.add(u=True)
    commit = repo.index.commit(commit_message)
    io.info('commit message: "%s"' % commit_message)
    io.info('created commit: %s' % commit.name_rev)
    return True


@git_failure
def push_commits(repo):
    io.info('pushing changes to %s' % repo.remotes.origin.refs[0].name)
    repo.remotes.origin.push()
    io.ok('successfully pushed changes to %s' % repo.remotes.origin.refs[0].name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
mimiron.py

usage:
    mim bump <service> <env> [--latest] [--no-push]
    mim status <env>

commands:
    bump          bumps the <service> with an image <artifact>
    status        shows the currently used artifact id for <env>

arguments:
    <artifact>    the deployment artifact we are pushing (e.g. Docker image/AMI)
    <service>     the application we're targeting
    <env>         the environment we want to change

options:
    --no-push     make local changes without pushing to remote
    --latest      use the latest artifact when updating a service

    -h --help     shows this
    -v --version  shows version
"""
from __future__ import print_function

from . import __version__
from docopt import docopt

import config
from .core import io
from .core.commands import bump
from .core.commands import status

from .domain import BaseMimironException
from .domain.commands import UnexpectedCommand


def _parse_user_input(args):
    env = args['<env>']

    if args['bump']:
        return bump.Bump(
            env=env,
            service=args['<service>'],
            is_latest=args['--latest'],
            should_push=not args['--no-push']
        )
    if args['status']:
        return status.Status(env=env)
    raise UnexpectedCommand


def main():
    args = docopt(__doc__, version=__version__)
    try:
        config.validate()
        config.post_process()

        _parse_user_input(args)
    except KeyboardInterrupt:
        pass
    except BaseMimironException as e:
        io.err(e)
        exit(1)
    exit(0)


if __name__ == '__main__':
    main()
